# -*- coding: utf-8 -*-
# Zilliqa Python Library
# Copyright (C) 2019  Gully Chen
# MIT License
"""
pyzil.common.utils
~~~~~~~~~~~~

common utils

:copyright: (c) 2019 by Gully Chen.
:license: MIT License, see LICENSE for more details.
"""

import string
import secrets
from typing import Union, Optional


TOKEN_NUM_BYTES = 32
TOKEN_STR_LENGTH = TOKEN_NUM_BYTES * 2


def randbelow(exclusive_upper_bound: int) -> int:
    """Return a random int in the range [0, n)."""
    return secrets.randbelow(exclusive_upper_bound)


def rand_string(n_len: int) -> str:
    """Return a random string containing n_len chars."""
    alphabet = string.ascii_letters + string.digits
    return "".join(secrets.choice(alphabet) for _ in range(n_len))


def rand_bytes(n_bytes: int=TOKEN_NUM_BYTES) -> bytes:
    """Return a random binary containing n_bytes bytes."""
    return secrets.token_bytes(n_bytes)


def rand_hex_str(n_len: int=TOKEN_STR_LENGTH, prefix: str="") -> str:
    """Return a random string containing n_len chars."""
    return prefix + (secrets.token_hex(n_len // 2 + 1)[:n_len])


def hex_str_to_bytes(str_hex: str) -> bytes:
    """Convert hex string to bytes."""
    str_hex = str_hex.lower()
    if str_hex.startswith("0x"):
        str_hex = str_hex[2:]
    if len(str_hex) & 1:
        str_hex = "0" + str_hex
    return bytes.fromhex(str_hex)


def bytes_to_hex_str(bytes_hex: bytes, prefix="") -> str:
    """Convert bytes to hex string."""
    return prefix + bytes_hex.hex()


def int_to_bytes(i: int, n_bytes: Optional[int]=TOKEN_NUM_BYTES,
                 byteorder="big") -> bytes:
    """Convert int to bytes."""
    if n_bytes is None:
        n_bytes = (i.bit_length() + 7) // 8 or 1
    return i.to_bytes(length=n_bytes, byteorder=byteorder)


def bytes_to_int(bytes_hex: bytes, byteorder="big") -> int:
    """Convert bytes to int."""
    return int.from_bytes(bytes_hex, byteorder=byteorder)


def int_to_hex_str(i: int, n_bytes: Optional[int]=TOKEN_NUM_BYTES,
                   prefix="", byteorder="big") -> str:
    """Convert int to hex string."""
    return bytes_to_hex_str(int_to_bytes(i, n_bytes, byteorder), prefix=prefix)


def hex_str_to_int(str_hex: str, byteorder="big") -> int:
    """Convert hex string to int."""
    return bytes_to_int(hex_str_to_bytes(str_hex), byteorder=byteorder)


def ensure_bytes(str_or_bytes: Union[str, bytes],
                 encoding="utf-8", errors="strict") -> bytes:
    """Convert input to bytes if it's str."""
    if isinstance(str_or_bytes, str):
        return str_or_bytes.encode(encoding=encoding, errors=errors)

    if not isinstance(str_or_bytes, bytes):
        raise TypeError("Type bytes required")

    return str_or_bytes



# -*- coding: utf-8 -*-
# Zilliqa Python Library
# Copyright (C) 2019  Gully Chen
# MIT License
"""
pyzil.account
~~~~~~~~~~~~

Zilliqa Account

:copyright: (c) 2019 by Gully Chen.
:license: MIT License, see LICENSE for more details.
"""

from typing import List, Union, Optional
from collections import namedtuple
from concurrent import futures
from concurrent.futures import ThreadPoolExecutor

from pyzil.crypto import zilkey
from pyzil.zilliqa.api import APIError
from pyzil.zilliqa.chain import active_chain
from pyzil.zilliqa.units import Qa, Zil


BatchTransfer = namedtuple("BatchTransfer", ["to_addr", "zils"])


class Account:
    """Zilliqa Account"""

    _min_gas = None

    def __init__(self, address=None, public_key=None, private_key=None):
        if address is None and public_key is None and private_key is None:
            raise ValueError("missing argument")

        self.address = None
        if address is not None:
            if not zilkey.is_valid_address(address):
                raise ValueError("invalid address")
            self.address = zilkey.to_valid_address(address)

        self.zil_key = None
        if public_key or private_key:
            self.zil_key = zilkey.ZilKey(public_key=public_key, private_key=private_key)

            if self.address is not None:
                if self.zil_key.address != self.address:
                    raise ValueError("mismatch address and zilkey")
            self.address = self.zil_key.address

    def __str__(self):
        return "<Account: {}>".format(self.address)

    def __eq__(self, other):
        if self.zil_key is None and other.zil_key is None:
            return self.address == other.address

        if self.zil_key is None or other.zil_key is None:
            return False

        return self.zil_key == other.zil_key

    @property
    def checksum_address(self) -> str:
        """Return str of checksum address."""
        return zilkey.to_checksum_address(self.address)

    @property
    def public_key(self) -> Optional[str]:
        """Return str of public key."""
        return self.zil_key and self.zil_key.keypair_str.public

    @property
    def private_key(self) -> Optional[str]:
        """Return str of private key."""
        return self.zil_key and self.zil_key.keypair_str.private

    @property
    def keypair(self) -> Optional[zilkey.KeyPair]:
        """Return keypair."""
        return self.zil_key and self.zil_key.keypair_str

    @classmethod
    def from_zilkey(cls, zil_key: zilkey.ZilKey) -> "Account":
        """Init account from a ZilKey instance."""
        return cls(private_key=zil_key.encoded_private_key)

    @classmethod
    def generate(cls) -> "Account":
        """Generate new account."""
        zil_key = zilkey.ZilKey.generate_new()
        return cls.from_zilkey(zil_key)

    @classmethod
    def from_mykey_txt(cls, key_file="mykey.txt") -> "Account":
        """Load account from mykey.txt."""
        zil_key = zilkey.ZilKey.load_mykey_txt(key_file)
        return cls.from_zilkey(zil_key)

    @classmethod
    def from_keystore(cls, password: str, keystore_file: str) -> "Account":
        """Load account from keystore json file."""
        zil_key = zilkey.ZilKey.load_keystore(password, keystore_file)
        return cls.from_zilkey(zil_key)

    @classmethod
    def get_min_gas_price(cls, refresh=False) -> int:
        if refresh or cls._min_gas is None:
            cls._min_gas = int(active_chain.api.GetMinimumGasPrice())
        return cls._min_gas

    def get_balance_nonce(self) -> dict:
        """Return raw response of GetBalance."""
        resp = {"balance": 0, "nonce": 0}
        try:
            resp = active_chain.api.GetBalance(self.address)
        except APIError as e:
            if str(e) != "Account is not created":
                raise e
        return resp

    def get_balance(self) -> Zil:
        """Return account balance in Zil."""
        resp = self.get_balance_nonce()
        return Qa(resp["balance"]).toZil()

    def get_balance_qa(self) -> Qa:
        resp = self.get_balance_nonce()
        return Qa(resp["balance"])

    def get_nonce(self) -> int:
        """Return account nonce."""
        resp = self.get_balance_nonce()
        return int(resp["nonce"])

    def transfer(self, to_addr: str,
                 zils: Union[str, float, Zil, Qa],
                 nonce: Optional[int]=None,
                 gas_price: Optional[int]=None, gas_limit=1,
                 code="", data="", priority=False):
        """Transfer zils to another address."""
        if not self.zil_key or not self.zil_key.encoded_private_key:
            raise RuntimeError("can not create transaction without private key")

        to_addr = zilkey.to_checksum_address(to_addr)
        if not to_addr:
            raise ValueError("invalid to address")

        if isinstance(zils, Qa):
            amount = zils
        else:
            if not isinstance(zils, Zil):
                zils = Zil(zils)
            amount = zils.toQa()

        if gas_price is None:
            gas_price = self.get_min_gas_price(refresh=False)

        if nonce is None:
            resp = self.get_balance_nonce()
            if amount > Qa(resp["balance"]):
                raise ValueError("insufficient balance to send")
            nonce = resp["nonce"] + 1

        params = active_chain.build_transaction_params(
            self.zil_key, to_addr,
            amount, nonce,
            gas_price, gas_limit,
            code, data, priority
        )

        txn_info = active_chain.api.CreateTransaction(params)
        return txn_info

    def transfer_batch(self, batch: List[BatchTransfer],
                       gas_price: Optional[int]=None, gas_limit=1,
                       max_workers=200, timeout=None):
        """Batch Transfer zils to addresses."""
        if not self.zil_key or not self.zil_key.encoded_private_key:
            raise RuntimeError("can not create transaction without private key")

        if gas_price is None:
            gas_price = self.get_min_gas_price(refresh=False)

        resp = self.get_balance_nonce()
        batch_nonce = resp["nonce"] + 1

        txn_params = []
        for to_addr, zils in batch:
            to_addr = zilkey.to_checksum_address(to_addr)
            if not to_addr:
                raise ValueError("invalid to address")

            if isinstance(zils, Qa):
                amount = zils
            else:
                if not isinstance(zils, Zil):
                    zils = Zil(zils)
                amount = zils.toQa()

            params = active_chain.build_transaction_params(
                self.zil_key, to_addr,
                amount, batch_nonce,
                gas_price, gas_limit
            )
            txn_params.append(params)
            batch_nonce += 1

        def create_txn(p):
            try:
                txn_info = active_chain.api.CreateTransaction(p)
            except Exception as e:
                print("Error in CreateTransaction: {}".format(e))
                txn_info = None
            return txn_info

        txn_results = []
        with ThreadPoolExecutor(max_workers=max_workers) as pool:
            all_tasks = [pool.submit(create_txn, p) for p in txn_params]

            for future in futures.as_completed(all_tasks, timeout=timeout):
                try:
                    txn_results.append(future.result())
                except Exception as e:
                    print("Error: {}".format(e))
                    txn_results.append(None)

        return txn_results

    @classmethod
    def wait_txn_confirm(cls, txn_id, timeout=300, sleep=20):
        return active_chain.wait_txn_confirm(txn_id, timeout=timeout, sleep=sleep)




# -*- coding: utf-8 -*-
# Zilliqa Python Library
# Copyright (C) 2019  Gully Chen
# MIT License
"""
pyzil.zilliqa.chain
~~~~~~~~~~~~

Zilliqa Blockchain.

:copyright: (c) 2019 by Gully Chen.
:license: MIT License, see LICENSE for more details.
"""

import time
from typing import Union, Optional

from pyzil.common import utils
from pyzil.common.local import LocalProxy
from pyzil.zilliqa.api import ZilliqaAPI, APIError
from pyzil.crypto.zilkey import is_valid_checksum_address, ZilKey
from pyzil.zilliqa.proto import messages_pb2 as pb2


class BlockChainError(Exception):
    pass


_active_chain: Optional["BlockChain"] = None


def get_active_chain() -> "BlockChain":
    if _active_chain is None:
        raise BlockChainError("active chain is not set, please call set_active_chain first")
    return _active_chain


def set_active_chain(chain: Optional["BlockChain"]) -> None:
    global _active_chain
    _active_chain = chain


active_chain = LocalProxy(get_active_chain)


class BlockChain:
    """Zilliqa Block Chain."""
    def __init__(self, api_url: str, version: Union[str, int], network_id: Union[str, int]):
        self.api_url = api_url
        self.version = version
        self.network_id = network_id
        self.api = ZilliqaAPI(endpoint=self.api_url)

    def __str__(self):
        return "<BlockChain: {}>".format(self.api_url)

    def build_transaction_params(self, zil_key: ZilKey, to_addr: str,
                                 amount: Union[str, int], nonce: Union[str, int],
                                 gas_price: Union[str, int], gas_limit: Union[str, int],
                                 code="", data="", priority=False):

        if not is_valid_checksum_address(to_addr):
            raise ValueError("invalid checksum address")

        txn_proto = pb2.ProtoTransactionCoreInfo()
        txn_proto.version = self.version
        txn_proto.nonce = int(nonce)
        txn_proto.toaddr = utils.hex_str_to_bytes(to_addr)
        txn_proto.senderpubkey.data = zil_key.keypair_bytes.public
        txn_proto.amount.data = utils.int_to_bytes(int(amount), n_bytes=16)
        txn_proto.gasprice.data = utils.int_to_bytes(int(gas_price), n_bytes=16)
        txn_proto.gaslimit = int(gas_limit)
        if code:
            txn_proto.code = code.encode("utf-8")
        if data:
            txn_proto.data = data.encode("utf-8")

        data_to_sign = txn_proto.SerializeToString()
        signature = zil_key.sign_str(data_to_sign)
        # assert zil_key.verify(signature, data_to_sign)

        params = {
            "version": txn_proto.version,
            "nonce": txn_proto.nonce,
            "toAddr": to_addr,
            "amount": str(amount),
            "pubKey": zil_key.keypair_str.public,
            "gasPrice": str(gas_price),
            "gasLimit": str(gas_limit),
            "code": code or None,
            "data": data or None,
            "signature": signature,
            "priority": priority,
        }
        return params

    def wait_txn_confirm(self, txn_id, timeout=60, sleep=5):
        start = time.time()
        while time.time() - start <= timeout:
            try:
                return self.api.GetTransaction(txn_id)
            except APIError as e:
                print("Retry GetTransaction: {}".format(e))
                time.sleep(sleep)
        return None


TestNet = BlockChain("https://dev-api.zilliqa.com/",
                     version=21823489, network_id=333)


MainNet = BlockChain("https://api.zilliqa.com/",
                     version=65537, network_id=1)


if "__main__" == __name__:
    print(TestNet.api.GetCurrentMiniEpoch())
    print(TestNet.api.GetCurrentDSEpoch())
    print(TestNet.api.GetBalance("b50c2404e699fd985f71b2c3f032059f13d6543b"))
    print(TestNet.api.GetBalance("4BAF5faDA8e5Db92C3d3242618c5B47133AE003C"))
    print(TestNet.api.GetBalance("4BAF5faDA8e5Db92C3d3242618c5B47133AE003C"))



# -*- coding: utf-8 -*-
# Zilliqa Python Library
# Copyright (C) 2019  Gully Chen
# MIT License
"""
pyzil.crypto.schnorr
~~~~~~~~~~~~

Zilliqa schnorr signature

:copyright: (c) 2019 by Gully Chen.
:license: MIT License, see LICENSE for more details.
"""

import secrets
import hashlib
from typing import Optional

from fastecdsa import keys
from fastecdsa import point
from fastecdsa import curve


CURVE = curve.secp256k1
CURVE_BITS = 256
ENCODED_SIZE = CURVE_BITS // 8

SECP256K1_TAG_PUBKEY_EVEN = b"\x02"
SECP256K1_TAG_PUBKEY_ODD = b"\x03"
SECP256K1_TAG_PUBKEY_UNCOMPRESSED = b"\x04"


def gen_private_key() -> int:
    """Generate a private key."""
    return keys.gen_private_key(CURVE)


def get_public_key(private_key: int) -> point.Point:
    """Get public key from a private key."""
    return keys.get_public_key(private_key, CURVE)


def encode_signature(r: int, s: int, size=ENCODED_SIZE) -> bytes:
    """encode signature to bytes."""
    r = r.to_bytes(size, "big")
    s = s.to_bytes(size, "big")
    return r + s


def decode_signature(signature: bytes) -> (int, int):
    """decode signature bytes to integer (r, s)."""
    size = len(signature) // 2
    r = int.from_bytes(signature[0:size], "big")
    s = int.from_bytes(signature[size:], "big")
    return r, s


def encode_public(x: int, y: int, compressed=True) -> bytes:
    """encode public key to bytes."""
    if compressed:
        tag = SECP256K1_TAG_PUBKEY_ODD if (y & 0x01) else SECP256K1_TAG_PUBKEY_EVEN
        return tag + x.to_bytes(ENCODED_SIZE, "big")
    else:
        tag = SECP256K1_TAG_PUBKEY_UNCOMPRESSED
        return tag + x.to_bytes(ENCODED_SIZE, "big") + y.to_bytes(ENCODED_SIZE, "big")


def decode_public(pub_key: bytes) -> point.Point:
    """decode public key from bytes to Point"""
    try:
        if len(pub_key) == 33:
            # compressed format
            y_odd = pub_key[0:1] == SECP256K1_TAG_PUBKEY_ODD
            x = int.from_bytes(pub_key[1:], "big")

            y_y = pow(x, 3) + CURVE.a * x + CURVE.b
            y = mod_sqrt(y_y, CURVE.p, y_odd)

        elif len(pub_key) == 65:
            tag = pub_key[0:1]
            assert tag == SECP256K1_TAG_PUBKEY_UNCOMPRESSED

            x = int.from_bytes(pub_key[1:ENCODED_SIZE + 1], "big")
            y = int.from_bytes(pub_key[ENCODED_SIZE + 1:], "big")

        else:
            raise NotImplementedError

        return point.Point(x, y, CURVE)
    except:
        raise ValueError("The public key could not be parsed or is invalid")


def mod_sqrt(n: int, p: int, is_odd: bool) -> int:
    """ Find Square Root under Modulo p
    Given a number 'n' and a prime 'p', find square root of n under modulo p if it exists.
    https://www.geeksforgeeks.org/find-square-root-under-modulo-p-set-1-when-p-is-in-form-of-4i-3/
    """
    n %= p
    y = pow(n, (p + 1) // 4, p)

    y_odd = bool(y & 0x01)
    if y_odd != is_odd:
        y = p - y

    assert (y * y) % p == n
    return y


def sign(bytes_msg: bytes, bytes_private: bytes, retries=10) -> Optional[bytes]:
    """sign bytes message with private key."""
    for i in range(retries):
        k = secrets.randbelow(CURVE.q)
        if k == 0:
            continue
        signature = sign_with_k(bytes_msg, bytes_private, k)
        if signature:
            return signature
    return None


def sign_with_k(bytes_msg: bytes,
                bytes_private: bytes,
                k: int) -> Optional[bytes]:
    """Sign bytes using Zilliqa schnorr signature algorithm."""
    assert isinstance(bytes_msg, bytes)
    private_key = int.from_bytes(bytes_private, "big")

    order = CURVE.q

    Q = CURVE.G * k
    bytes_Q_x = encode_public(Q.x, Q.y)

    pub_key = keys.get_public_key(private_key, CURVE)
    bytes_pub_x = encode_public(pub_key.x, pub_key.y)

    hasher = hashlib.sha256()
    hasher.update(bytes_Q_x + bytes_pub_x + bytes_msg)
    r = hasher.digest()
    r = int.from_bytes(r, "big") % order
    s = (k - r * private_key) % order
    if r == 0 or s == 0:
        return None

    return encode_signature(r, s)


def verify(bytes_msg: bytes,
           signature: bytes,
           bytes_public: bytes) -> bool:
    """verify schnorr signature with public key."""
    assert isinstance(bytes_msg, bytes)
    assert isinstance(signature, bytes)
    pub_key = decode_public(bytes_public)

    r, s = decode_signature(signature)
    n, G = CURVE.q, CURVE.G

    if not s or s >= n:
        return False
    if not r or r >= pow(2, CURVE_BITS):
        return False

    sG = s * G
    rW = r * pub_key
    Q = sG + rW

    bytes_Q_x = encode_public(Q.x, Q.y)
    bytes_pub_x = encode_public(pub_key.x, pub_key.y)

    hasher = hashlib.sha256()
    hasher.update(bytes_Q_x + bytes_pub_x + bytes_msg)
    v = hasher.digest()
    v = int.from_bytes(v, "big")
    v = v % n

    return v == r



# -*- coding: utf-8 -*-
# Zilliqa Python Library
# Copyright (C) 2019  Gully Chen
# MIT License
"""
pyzil.crypto.zilkey
~~~~~~~~~~~~

Zilliqa Key

:copyright: (c) 2019 by Gully Chen.
:license: MIT License, see LICENSE for more details.
"""

import json
import uuid
from typing import Union, Optional
from collections import namedtuple

from pyzil.common import utils
from pyzil.crypto import tools, schnorr


# zilliqa address takes the last 20 bytes from hash digest of public key
ADDRESS_NUM_BYTES = 20
ADDRESS_STR_LENGTH = ADDRESS_NUM_BYTES * 2


def is_valid_address(address: str) -> bool:
    """Return True if address is valid."""
    if address.lower().startswith("0x"):
        address = address[2:]
    if len(address) != ADDRESS_STR_LENGTH:
        return False
    # noinspection PyBroadException
    try:
        utils.hex_str_to_int(address)
    except Exception:
        return False
    return True


def to_valid_address(address: str) -> Optional[str]:
    """Return lower case address if address is valid."""
    if not is_valid_address(address):
        return None
    address = address.lower()
    if address.startswith("0x"):
        address = address[2:]
    return address


def to_checksum_address(address: str, prefix="0x") -> Optional[str]:
    """Convert address to checksum address."""
    if not is_valid_address(address):
        return None

    address = address.lower().replace("0x", "")
    address_bytes = utils.hex_str_to_bytes(address)
    v = utils.bytes_to_int(tools.hash256_bytes(address_bytes))

    checksum_address = prefix
    for i, c in enumerate(address):
        if not c.isdigit():
            if v & (1 << 255 - 6 * i):
                c = c.upper()
            else:
                c = c.lower()
        checksum_address += c

    return checksum_address


def is_valid_checksum_address(address: str) -> bool:
    """Return True if address is valid checksum address."""
    if not is_valid_address(address):
        return False
    return to_checksum_address(address) == address


KeyPair = namedtuple("KeyPair", ["public", "private"])


class ZilKey:
    """ Zilliqa Key """
    def __init__(self, public_key=None, private_key=None):
        assert public_key or private_key, "public or private key is required"
        if isinstance(public_key, str):
            public_key = utils.hex_str_to_bytes(public_key)
        if isinstance(private_key, str):
            private_key = utils.hex_str_to_bytes(private_key)

        self._bytes_public = public_key
        self._bytes_private = private_key

        # the _pub_key is a Point on curve
        self._public_key = None
        # the private_key is big integer less than curve order
        self._private_key = None

        self._generate_keys()

    def _generate_keys(self):
        if self._bytes_private:
            self._private_key = utils.bytes_to_int(self._bytes_private)
            assert self._private_key < schnorr.CURVE.q
        if self._bytes_public:
            self._public_key = schnorr.decode_public(self._bytes_public)

        if self._private_key and self._public_key:
            _pub_key = schnorr.get_public_key(self._private_key)
            assert _pub_key == self._public_key, "public/private key mismatch"

        # generate public key from private key
        if self._private_key and not self._public_key:
            self._public_key = schnorr.get_public_key(self._private_key)

    @property
    def encoded_public_key(self):
        """bytes of public key."""
        return schnorr.encode_public(self._public_key.x, self._public_key.y)

    @property
    def encoded_private_key(self):
        """bytes of private key."""
        return self._private_key and utils.int_to_bytes(self._private_key)

    @property
    def keypair_bytes(self) -> KeyPair:
        """bytes of key pair."""
        return KeyPair(self.encoded_public_key, self.encoded_private_key)

    @property
    def keypair_str(self) -> KeyPair:
        """hex string of key pair."""
        str_pub = utils.bytes_to_hex_str(self.encoded_public_key)
        str_private = self._private_key and utils.int_to_hex_str(self._private_key)
        return KeyPair(str_pub, str_private)

    @property
    def address(self) -> str:
        addr_bytes = tools.hash256_bytes(self.keypair_bytes.public)
        return utils.bytes_to_hex_str(addr_bytes)[-ADDRESS_STR_LENGTH:]

    @property
    def checksum_address(self) -> str:
        return to_checksum_address(self.address)

    def __str__(self):
        return str(self.keypair_str)

    def __eq__(self, other):
        return self._public_key == other._public_key and self._private_key == other._private_key

    # Zilliqa schnorr signature
    def sign(self, message: bytes) -> bytes:
        """Sign bytes message with private key, return bytes"""
        if not self._private_key:
            raise RuntimeError("missing private key")

        message = utils.ensure_bytes(message)

        return schnorr.sign(message, self.keypair_bytes.private)

    def sign_str(self, message: str) -> str:
        """Sign bytes message with private key, return hex string"""
        message = utils.ensure_bytes(message)
        return utils.bytes_to_hex_str(self.sign(message))

    def verify(self, signature: Union[str, bytes], message: Union[str, bytes]) -> bool:
        """Verify signature with public key."""
        if isinstance(signature, str):
            signature = utils.hex_str_to_bytes(signature)

        message = utils.ensure_bytes(message)

        return schnorr.verify(message, signature, self.keypair_bytes.public)

    @classmethod
    def generate_new(cls):
        """Generate new zilliqa key"""
        zil_key = cls(private_key=utils.int_to_bytes(schnorr.gen_private_key()))
        return zil_key

    @classmethod
    def load_mykey_txt(cls, key_file="mykey.txt"):
        """Load Zilliqa key from mykey.txt."""
        with open(key_file, "r") as f:
            str_pub, str_private = f.read().split()
            return ZilKey(public_key=str_pub, private_key=str_private)

    def save_mykey_txt(self, key_file="mykey.txt"):
        """Save key to mykey.txt."""
        with open(key_file, "w") as f:
            str_pub, str_private = self.keypair_str
            f.write(str_pub.upper() + " " + str_private.upper())

    @classmethod
    def load_keystore(cls, password: str, keystore_file):
        """Load Zilliqa key from keystore json file."""
        if hasattr(keystore_file, "read"):
            keystore = json.load(keystore_file)
        else:
            with open(keystore_file) as f:
                keystore = json.load(f)

        checksum_address = to_checksum_address(keystore["address"])
        ciphertext = utils.hex_str_to_bytes(keystore["crypto"]["ciphertext"])
        iv = utils.hex_str_to_bytes(keystore["crypto"]["cipherparams"]["iv"])

        kdf_method = keystore["crypto"]["kdf"]
        kdfparams = keystore["crypto"]["kdfparams"]

        derived_key = tools.gen_derived_key(password, kdf_method, kdfparams)

        cipher = keystore["crypto"]["cipher"].encode()
        message = derived_key[16:32] + ciphertext + iv + cipher
        mac_generated = tools.hmac_hash256(derived_key, msg=message)

        mac_from_file = utils.hex_str_to_bytes(keystore["crypto"]["mac"])
        if not tools.compare_digest(mac_generated, mac_from_file):
            raise ValueError("Invalid password or keystore file")

        key = derived_key[:16]
        private_key = tools.aes_ctr_decrypt(key, iv, ciphertext)
        zilkey = cls(private_key=private_key)
        if not tools.compare_digest(checksum_address, zilkey.checksum_address):
            raise ValueError("Invalid keystore file, address mismatch")
        return zilkey

    def save_keysotre(self, password: str, kdf_method: str="pbkdf2",
                      keystore_file=None) -> dict:
        """Save Zilliqa key to keystore, format details on
        https://github.com/ethereum/wiki/wiki/Web3-Secret-Storage-Definition.
        """
        address = self.address
        cipher = "aes-128-ctr"
        salt = utils.rand_bytes(32)
        iv = utils.rand_bytes(16)
        kdfparams = {
            "salt": utils.bytes_to_hex_str(salt),
            "n": 8192,
            "c": 262144,
            "r": 8,
            "p": 1,
            "dklen": 32
        }
        derived_key = tools.gen_derived_key(password, kdf_method, kdfparams)

        key = derived_key[:16]
        ciphertext = tools.aes_ctr_encrypt(key, iv,
                                           ciphertext=self.encoded_private_key)
        cipher_bytes = cipher.encode()
        message = derived_key[16:32] + ciphertext + iv + cipher_bytes
        mac_generated = tools.hmac_hash256(derived_key, msg=message)

        keystore = {
            "address": address,
            "crypto": {
                "cipher": cipher,
                "cipherparams": {
                    "iv": utils.bytes_to_hex_str(iv),
                },
                "ciphertext": utils.bytes_to_hex_str(ciphertext),
                "kdf": kdf_method,
                "kdfparams": kdfparams,
                "mac": utils.bytes_to_hex_str(mac_generated),
            },
            "id": str(uuid.uuid4()),
            "version": 3,
        }

        if keystore_file is None:
            return keystore
        if hasattr(keystore_file, "write"):
            json.dump(keystore, keystore_file)
        else:
            with open(keystore_file, "w") as f:
                json.dump(keystore, f)
        return keystore



# -*- coding: utf-8 -*-
# Zilliqa Python Library
# Copyright (C) 2019  Gully Chen
# MIT License
"""
pyzil.crypto
~~~~~~~~~~~~

This module implements the cryptographic methods

:copyright: (c) 2019 by Gully Chen.
:license: MIT License, see LICENSE for more details.
"""

__all__ = ["zilkey", "ZilKey", "KeyPair"]

from . import zilkey
from .zilkey import ZilKey, KeyPair



# -*- coding: utf-8 -*-
# Zilliqa Python Library
# Copyright (C) 2019  Gully Chen
# MIT License
"""
pyzil.crypto.tools
~~~~~~~~~~~~

crypto tools

:copyright: (c) 2019 by Gully Chen.
:license: MIT License, see LICENSE for more details.
"""

import hmac
import hashlib
from typing import Union

from Crypto.Cipher import AES

from pyzil.common import utils


def hash256_bytes(*bytes_hex, encoding="utf-8") -> bytes:
    """Return hash256 digest bytes."""
    m = hashlib.sha256()
    for b in bytes_hex:
        if isinstance(b, str):
            b = b.encode(encoding=encoding)
        m.update(b)
    return m.digest()


def hmac_hash256(key: bytes, msg: bytes) -> bytes:
    """Return bytes of hmac using hash256."""
    h = hmac.new(key, msg, digestmod=hashlib.sha256)
    return h.digest()


def compare_digest(a, b) -> bool:
    return hmac.compare_digest(a, b)


def gen_derived_key(password: Union[str, bytes],
                    kdf_method: str, params: dict) -> bytes:
    """Generate derived key bytes using pbkdf2 or scrypt."""
    if isinstance(password, str):
        password = password.encode()
    salt = params["salt"]
    if isinstance(salt, str):
        salt = utils.hex_str_to_bytes(salt)

    dklen = params["dklen"]
    if kdf_method == "pbkdf2":
        count = params["c"]
        return hashlib.pbkdf2_hmac(
            "sha256", password, salt, count, dklen=dklen
        )
    elif kdf_method == "scrypt":
        return hashlib.scrypt(
            password, salt,
            params["n"], params["r"], params["p"],
            dklen=dklen
        )
    else:
        raise NotImplemented("unsupported kdf method")


def aes_ctr_decrypt(key: bytes, initial_value: bytes,
                    ciphertext: bytes, nonce: bytes=b"") -> bytes:
    cipher = AES.new(key, AES.MODE_CTR,
                     nonce=nonce, initial_value=initial_value)
    return cipher.decrypt(ciphertext)


def aes_ctr_encrypt(key: bytes, initial_value: bytes,
                    ciphertext: bytes, nonce: bytes=b"") -> bytes:
    cipher = AES.new(key, AES.MODE_CTR,
                     nonce=nonce, initial_value=initial_value)
    return cipher.encrypt(ciphertext)



# -*- coding: utf-8 -*-
# Zilliqa Python Library
# Copyright (C) 2019  Gully Chen
# MIT License
"""
pyzil.pow
~~~~~~~~~~~~

This module implements Zilliqa PoW methods

:copyright: (c) 2019 by Gully Chen.
:license: MIT License, see LICENSE for more details.
"""


from typing import List, Tuple, Optional, Union
from collections import OrderedDict

from pyzil.common import utils

from pyethash import (
    EPOCH_LENGTH,
    hashimoto_light,
    get_seedhash,
    mkcache_bytes,
)
from eth_hash.auto import keccak


MAX_EPOCH = 2048


def block_num_to_seed(block_number: int) -> bytes:
    """DS block number to seed hash."""
    return get_seedhash(block_number)


def seed_to_epoch_num(seed: bytes) -> int:
    """Seed to epoch number."""
    for epoch in range(MAX_EPOCH):
        block_num = epoch * EPOCH_LENGTH + 1
        calc_seed = block_num_to_seed(block_num)
        if seed == calc_seed:
            return epoch
    raise ValueError("epoch number out of range, max 2048")


def seed_to_block_num(seed: bytes) -> int:
    """Seed to DS block number."""
    return seed_to_epoch_num(seed) * EPOCH_LENGTH


ZERO_MASK = [0xFF, 0x7F, 0x3F, 0x1F, 0x0F, 0x07, 0x03, 0x01]


def difficulty_to_boundary(difficulty: int) -> bytes:
    """Zilliqa difficulty level to boundary."""
    boundary = bytearray(b"\xFF" * 32)
    n_bytes_to_zero = difficulty // 8
    n_bits_to_zero = difficulty % 8

    boundary[0:n_bytes_to_zero] = b"\x00" * n_bytes_to_zero
    boundary[n_bytes_to_zero] = ZERO_MASK[n_bits_to_zero]

    return bytes(boundary)


def boundary_to_difficulty(boundary) -> int:
    """Boundary to Zilliqa difficulty level."""
    if isinstance(boundary, str):
        boundary = utils.hex_str_to_bytes(boundary)

    difficulty = 0
    for b in memoryview(boundary):
        if b == 0x00:
            difficulty += 8
        else:
            difficulty += (8 - b.bit_length())
            break
    return difficulty


assert boundary_to_difficulty(difficulty_to_boundary(11)) == 11


def difficulty_to_boundary_divided(difficulty: int, n_divided: int=8,
                                   n_divided_start: int=32) -> bytes:
    """Zilliqa divided difficulty to boundary."""
    if difficulty < n_divided_start:
        return difficulty_to_boundary(difficulty)

    n_level = (difficulty - n_divided_start) // n_divided
    m_sub_level = (difficulty - n_divided_start) % n_divided
    difficulty_level = n_divided_start + n_level

    int_boundary = utils.bytes_to_int(difficulty_to_boundary(difficulty_level))
    boundary_change_step = (int_boundary >> 1) // n_divided

    int_boundary -= boundary_change_step * m_sub_level

    return utils.int_to_bytes(int_boundary, n_bytes=32)


def boundary_to_difficulty_divided(boundary, n_divided: int=8,
                                   n_divided_start: int=32) -> int:
    """Boundary to divided difficulty."""
    if isinstance(boundary, str):
        boundary = utils.hex_str_to_bytes(boundary)

    difficulty_level = boundary_to_difficulty(boundary)
    if difficulty_level < n_divided_start:
        return difficulty_level

    n_level = difficulty_level - n_divided_start

    int_cur_boundary = utils.bytes_to_int(boundary)
    int_cur_level = utils.bytes_to_int(difficulty_to_boundary(difficulty_level))

    step = (int_cur_level >> 1) // n_divided
    m_sub_level = (int_cur_level - int_cur_boundary) // step

    new_difficulty = n_divided_start + n_level * n_divided + m_sub_level
    return new_difficulty


assert boundary_to_difficulty_divided(difficulty_to_boundary_divided(31)) == 31
assert boundary_to_difficulty_divided(difficulty_to_boundary_divided(32)) == 32
assert boundary_to_difficulty_divided(difficulty_to_boundary_divided(49)) == 49
assert boundary_to_difficulty_divided(difficulty_to_boundary_divided(255)) == 255
assert boundary_to_difficulty_divided(difficulty_to_boundary(31)) == 31


def boundary_to_hashpower(boundary: Union[str, bytes]) -> int:
    """boundary to hashrate."""
    dividend = 0xffff000000000000000000000000000000000000000000000000000000000000
    if isinstance(boundary, str):
        return dividend // utils.hex_str_to_int(boundary)
    elif isinstance(boundary, bytes):
        return dividend // utils.bytes_to_int(boundary)
    raise TypeError("Type of boundary should be str or bytes")


def difficulty_to_hashpower(difficulty: int) -> int:
    """difficulty level to hashrate."""
    return boundary_to_hashpower(difficulty_to_boundary(difficulty))


def difficulty_to_hashpower_divided(difficulty: int, n_divided: int=8,
                                    n_divided_start: int=32) -> int:
    """divided difficulty to hashrate."""
    return boundary_to_hashpower(
        difficulty_to_boundary_divided(
            difficulty, n_divided=n_divided, n_divided_start=n_divided_start
        )
    )


def is_less_or_equal(hash_1: Union[str, bytes],
                     hash_2: Union[str, bytes]) -> bool:
    """check hash result."""
    if isinstance(hash_1, str):
        hash_1 = utils.hex_str_to_bytes(hash_1)
    if isinstance(hash_2, str):
        hash_2 = utils.hex_str_to_bytes(hash_2)
    assert isinstance(hash_1, bytes)
    assert isinstance(hash_2, bytes)

    return utils.bytes_to_int(hash_1) <= utils.bytes_to_int(hash_2)


# for pow verify
def verify_pow_work(block_number: int, header: bytes, mix_digest: bytes,
                    nonce: int, boundary: bytes) -> Optional[bytes]:
    """Return hash rate if it less than boundary."""
    calc_mix_digest, calc_result = pow_hash(block_number, header, nonce)

    if mix_digest != calc_mix_digest:
        return None

    ok = is_less_or_equal(calc_result, boundary)
    if not ok:
        return None
    return calc_result


# ethash settings
CACHE_MAX_ITEMS = 10
cache_seeds = bytearray(b"\x00" * 32)     # type: List[bytes]
cache_by_seed = OrderedDict()             # type: OrderedDict[bytes, bytearray]


def get_cache(block_number: int) -> bytes:
    """helper function for ethash."""
    while len(cache_seeds) <= block_number // EPOCH_LENGTH:
        cache_seeds.append(keccak(cache_seeds[-1]))
    seed = cache_seeds[block_number // EPOCH_LENGTH]
    if seed in cache_by_seed:
        c = cache_by_seed.pop(seed)  # pop and append at end
        cache_by_seed[seed] = c
        return c
    c = mkcache_bytes(block_number)
    cache_by_seed[seed] = c
    if len(cache_by_seed) > CACHE_MAX_ITEMS:
        cache_by_seed.popitem(last=False)  # remove last recently accessed
    return c


def pow_hash(block_number, header, nonce) -> Tuple[bytes, bytes]:
    """search for hash result using hashimoto_light."""
    cache_bytes = get_cache(block_number)
    hash_ret = hashimoto_light(block_number, cache_bytes, header, nonce)
    return hash_ret[b"mix digest"], hash_ret[b"result"]



# -*- coding: utf-8 -*-
# Zilliqa Python Library
# Copyright (C) 2019  Gully Chen
# MIT License

"""
Python Library Zilliqa APIs
~~~~~~~~~~~~~~~~~~~~~

pyzil is a Python library for Zilliqa APIs ( https://apidocs.zilliqa.com )
usage:

   >>> from pyzil.account import Account
   >>> from pyzil.zilliqa import chain
   >>> account = Account(address="b50c2404e699fd985f71b2c3f032059f13d6543b")
   >>> chain.set_active_chain(chain.TestNet)
   >>> account.get_balance()
   610.269 Zil
   >>> account.get_nonce()
   3
   >>> chain.set_active_chain(chain.MainNet)
   >>> account.get_balance()
   0 Zil

:copyright: (c) 2019 by Gully Chen.
:license: MIT License, see LICENSE for more details.
"""

version = "0.4.04"



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: messages.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='messages.proto',
  package='ZilliqaMessage',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n\x0emessages.proto\x12\x0eZilliqaMessage\"\x19\n\tByteArray\x12\x0c\n\x04\x64\x61ta\x18\x01 \x02(\x0c\"\x81\x02\n\x18ProtoTransactionCoreInfo\x12\x0f\n\x07version\x18\x01 \x01(\r\x12\r\n\x05nonce\x18\x02 \x01(\x04\x12\x0e\n\x06toaddr\x18\x03 \x01(\x0c\x12/\n\x0csenderpubkey\x18\x04 \x01(\x0b\x32\x19.ZilliqaMessage.ByteArray\x12)\n\x06\x61mount\x18\x05 \x01(\x0b\x32\x19.ZilliqaMessage.ByteArray\x12+\n\x08gasprice\x18\x06 \x01(\x0b\x32\x19.ZilliqaMessage.ByteArray\x12\x10\n\x08gaslimit\x18\x07 \x01(\x04\x12\x0c\n\x04\x63ode\x18\x08 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\t \x01(\x0c\"\x88\x01\n\x10ProtoTransaction\x12\x0e\n\x06tranid\x18\x01 \x01(\x0c\x12\x36\n\x04info\x18\x02 \x01(\x0b\x32(.ZilliqaMessage.ProtoTransactionCoreInfo\x12,\n\tsignature\x18\x03 \x01(\x0b\x32\x19.ZilliqaMessage.ByteArray\":\n\x17ProtoTransactionReceipt\x12\x0f\n\x07receipt\x18\x01 \x01(\x0c\x12\x0e\n\x06\x63umgas\x18\x02 \x01(\x04\"\x8e\x01\n\x1bProtoTransactionWithReceipt\x12\x35\n\x0btransaction\x18\x01 \x01(\x0b\x32 .ZilliqaMessage.ProtoTransaction\x12\x38\n\x07receipt\x18\x02 \x01(\x0b\x32\'.ZilliqaMessage.ProtoTransactionReceipt')
)




_BYTEARRAY = _descriptor.Descriptor(
  name='ByteArray',
  full_name='ZilliqaMessage.ByteArray',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='ZilliqaMessage.ByteArray.data', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=34,
  serialized_end=59,
)


_PROTOTRANSACTIONCOREINFO = _descriptor.Descriptor(
  name='ProtoTransactionCoreInfo',
  full_name='ZilliqaMessage.ProtoTransactionCoreInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='ZilliqaMessage.ProtoTransactionCoreInfo.version', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='nonce', full_name='ZilliqaMessage.ProtoTransactionCoreInfo.nonce', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='toaddr', full_name='ZilliqaMessage.ProtoTransactionCoreInfo.toaddr', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='senderpubkey', full_name='ZilliqaMessage.ProtoTransactionCoreInfo.senderpubkey', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='amount', full_name='ZilliqaMessage.ProtoTransactionCoreInfo.amount', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='gasprice', full_name='ZilliqaMessage.ProtoTransactionCoreInfo.gasprice', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='gaslimit', full_name='ZilliqaMessage.ProtoTransactionCoreInfo.gaslimit', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='code', full_name='ZilliqaMessage.ProtoTransactionCoreInfo.code', index=7,
      number=8, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='ZilliqaMessage.ProtoTransactionCoreInfo.data', index=8,
      number=9, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=62,
  serialized_end=319,
)


_PROTOTRANSACTION = _descriptor.Descriptor(
  name='ProtoTransaction',
  full_name='ZilliqaMessage.ProtoTransaction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='tranid', full_name='ZilliqaMessage.ProtoTransaction.tranid', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='info', full_name='ZilliqaMessage.ProtoTransaction.info', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='signature', full_name='ZilliqaMessage.ProtoTransaction.signature', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=322,
  serialized_end=458,
)


_PROTOTRANSACTIONRECEIPT = _descriptor.Descriptor(
  name='ProtoTransactionReceipt',
  full_name='ZilliqaMessage.ProtoTransactionReceipt',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='receipt', full_name='ZilliqaMessage.ProtoTransactionReceipt.receipt', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cumgas', full_name='ZilliqaMessage.ProtoTransactionReceipt.cumgas', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=460,
  serialized_end=518,
)


_PROTOTRANSACTIONWITHRECEIPT = _descriptor.Descriptor(
  name='ProtoTransactionWithReceipt',
  full_name='ZilliqaMessage.ProtoTransactionWithReceipt',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='transaction', full_name='ZilliqaMessage.ProtoTransactionWithReceipt.transaction', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='receipt', full_name='ZilliqaMessage.ProtoTransactionWithReceipt.receipt', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=521,
  serialized_end=663,
)

_PROTOTRANSACTIONCOREINFO.fields_by_name['senderpubkey'].message_type = _BYTEARRAY
_PROTOTRANSACTIONCOREINFO.fields_by_name['amount'].message_type = _BYTEARRAY
_PROTOTRANSACTIONCOREINFO.fields_by_name['gasprice'].message_type = _BYTEARRAY
_PROTOTRANSACTION.fields_by_name['info'].message_type = _PROTOTRANSACTIONCOREINFO
_PROTOTRANSACTION.fields_by_name['signature'].message_type = _BYTEARRAY
_PROTOTRANSACTIONWITHRECEIPT.fields_by_name['transaction'].message_type = _PROTOTRANSACTION
_PROTOTRANSACTIONWITHRECEIPT.fields_by_name['receipt'].message_type = _PROTOTRANSACTIONRECEIPT
DESCRIPTOR.message_types_by_name['ByteArray'] = _BYTEARRAY
DESCRIPTOR.message_types_by_name['ProtoTransactionCoreInfo'] = _PROTOTRANSACTIONCOREINFO
DESCRIPTOR.message_types_by_name['ProtoTransaction'] = _PROTOTRANSACTION
DESCRIPTOR.message_types_by_name['ProtoTransactionReceipt'] = _PROTOTRANSACTIONRECEIPT
DESCRIPTOR.message_types_by_name['ProtoTransactionWithReceipt'] = _PROTOTRANSACTIONWITHRECEIPT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ByteArray = _reflection.GeneratedProtocolMessageType('ByteArray', (_message.Message,), dict(
  DESCRIPTOR = _BYTEARRAY,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:ZilliqaMessage.ByteArray)
  ))
_sym_db.RegisterMessage(ByteArray)

ProtoTransactionCoreInfo = _reflection.GeneratedProtocolMessageType('ProtoTransactionCoreInfo', (_message.Message,), dict(
  DESCRIPTOR = _PROTOTRANSACTIONCOREINFO,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:ZilliqaMessage.ProtoTransactionCoreInfo)
  ))
_sym_db.RegisterMessage(ProtoTransactionCoreInfo)

ProtoTransaction = _reflection.GeneratedProtocolMessageType('ProtoTransaction', (_message.Message,), dict(
  DESCRIPTOR = _PROTOTRANSACTION,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:ZilliqaMessage.ProtoTransaction)
  ))
_sym_db.RegisterMessage(ProtoTransaction)

ProtoTransactionReceipt = _reflection.GeneratedProtocolMessageType('ProtoTransactionReceipt', (_message.Message,), dict(
  DESCRIPTOR = _PROTOTRANSACTIONRECEIPT,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:ZilliqaMessage.ProtoTransactionReceipt)
  ))
_sym_db.RegisterMessage(ProtoTransactionReceipt)

ProtoTransactionWithReceipt = _reflection.GeneratedProtocolMessageType('ProtoTransactionWithReceipt', (_message.Message,), dict(
  DESCRIPTOR = _PROTOTRANSACTIONWITHRECEIPT,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:ZilliqaMessage.ProtoTransactionWithReceipt)
  ))
_sym_db.RegisterMessage(ProtoTransactionWithReceipt)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Zilliqa Python Library
# Copyright (C) 2019  Gully Chen
# MIT License
"""
pyzil.zilliqa.units
~~~~~~~~~~~~

Zilliqa currencies units.

:copyright: (c) 2019 by Gully Chen.
:license: MIT License, see LICENSE for more details.
"""


class Units:
    prec = 12
    fmt = "{:.12f}"
    units = pow(10, prec)


class Qa(int):
    def __repr__(self):
        return "{} Qa".format(int(self))

    def toZil(self):
        zil = int(self) / Units.units
        return Zil(zil)

    @classmethod
    def fromZil(cls, zil):
        return Zil(zil).toQa()

    def __eq__(self, other):
        if isinstance(other, Zil):
            other = other.toQa()
        return int(self) == int(other)

    def __add__(self, other):
        """Return self + other"""
        if isinstance(other, Zil):
            other = other.toQa()
        return Qa(int(self) + int(other))

    def __radd__(self, other):
        """Return other + self"""
        if isinstance(other, Zil):
            other = other.toQa()
        return Qa(int(other) + int(self))

    def __sub__(self, other):
        """Return self - other"""
        if isinstance(other, Zil):
            other = other.toQa()
        return Qa(int(self) - int(other))

    def __rsub__(self, other):
        """Return other - self"""
        if isinstance(other, Zil):
            other = other.toQa()
        return Qa(int(other) - int(self))

    def __mul__(self, other):
        """Return self * other"""
        return Qa(int(self) * other)

    def __rmul__(self, other):
        """Return other * self"""
        return Qa(other * int(self))

    def __truediv__(self, other):
        """Return self / other"""
        return Qa(int(self) / other)

    def __rtruediv__(self, other):
        """Return other / self"""
        return Qa(other / int(self))

    def __floordiv__(self, other):
        """Return self // other"""
        return Qa(int(self) // other)

    def __rfloordiv__(self, other):
        """Return other // self"""
        return Qa(other // int(self))


class Zil(float):
    def __str__(self):
        fmt_str = Units.fmt.format(float(self))
        return fmt_str.rstrip("0").rstrip(".")

    def __repr__(self):
        return self.__str__() + " Zil"

    def __eq__(self, other):
        if not isinstance(other, Qa):
            if not isinstance(other, Zil):
                other = Zil(other)
            other = other.toQa()
        return self.toQa() == other

    def toQa(self):
        qa = float(self) * Units.units
        return Qa(int(qa))

    @classmethod
    def fromQa(cls, qa):
        return Qa(qa).toZil()

    def __add__(self, other):
        """Return self + other"""
        if isinstance(other, Qa):
            other = other.toZil()
        return Zil(float(self) + float(other))

    def __radd__(self, other):
        """Return other + self"""
        if isinstance(other, Qa):
            other = other.toZil()
        return Zil(float(other) + float(self))

    def __sub__(self, other):
        """Return self - other"""
        if isinstance(other, Qa):
            other = other.toZil()
        return Zil(float(self) - float(other))

    def __rsub__(self, other):
        """Return other - self"""
        if isinstance(other, Qa):
            other = other.toZil()
        return Zil(float(other) - float(self))

    def __mul__(self, other):
        """Return self * other"""
        return Zil(float(self) * other)

    def __rmul__(self, other):
        """Return other * self"""
        return Zil(other * float(self))

    def __truediv__(self, other):
        """Return self / other"""
        return Zil(float(self) / other)

    def __rtruediv__(self, other):
        """Return other / self"""
        return Zil(other / float(self))

    def __floordiv__(self, other):
        """Return self // other"""
        return Zil(float(self) // other)

    def __rfloordiv__(self, other):
        """Return other // self"""
        return Zil(other // float(self))


if "__main__" == __name__:
    print(Zil(0) >= 0)



# -*- coding: utf-8 -*-
# Zilliqa Python Library
# Copyright (C) 2019  Gully Chen
# MIT License
"""
pyzil.zilliqa.api
~~~~~~~~~~~~

Json-RPC interface of Zilliqa APIs.

:copyright: (c) 2019 by Gully Chen.
:license: MIT License, see LICENSE for more details.
"""

from jsonrpcclient.exceptions import JsonRpcClientError
from jsonrpcclient.clients.http_client import HTTPClient


INVALID_PARAMS = "INVALID_PARAMS: Invalid method parameters (invalid name and/or type) recognised"


class APIError(Exception):
    pass


class ZilliqaAPI:
    """Json-RPC interface of Zilliqa APIs."""
    class APIMethod:
        def __init__(self, api: "ZilliqaAPI", method_name: str):
            self.api = api
            self.method_name = method_name

        def __call__(self, *params, **kwargs):
            resp = self.api.call(self.method_name, *params, **kwargs)
            return resp and resp.data and resp.data.result

    def __init__(self, endpoint: str):
        self.endpoint = endpoint
        self.api_client = HTTPClient(self.endpoint)

    def __getattr__(self, item: str):
        return ZilliqaAPI.APIMethod(self, method_name=item)

    def call(self, method_name: str, *params, **kwargs):

        def send_request(*_params):
            try:
                return self.api_client.request(
                    method_name, *_params,
                    trim_log_values=True, **kwargs
                )
            except JsonRpcClientError as _e:
                raise APIError(_e)

        try:
            return send_request(*params)
        except APIError as e:
            # fix for jsonrpcclient < 3.3.1
            if str(e) == INVALID_PARAMS:
                if len(params) == 1 and isinstance(params[0], (dict, list)):
                    params = (list(params),)
                    return send_request(*params)
            raise e


if "__main__" == __name__:
    _api = ZilliqaAPI("https://dev-api.zilliqa.com/")
    print(_api.GetCurrentMiniEpoch())
    print(_api.GetCurrentDSEpoch())
    print(_api.GetBalance("b50c2404e699fd985f71b2c3f032059f13d6543b"))
    print(_api.GetBalance("4BAF5faDA8e5Db92C3d3242618c5B47133AE003C"))
    print(_api.GetBalance("4BAF5faDA8e5Db92C3d3242618c5B47133AE003C"))




# -*- coding: utf-8 -*-
# Zilliqa Python API
# Copyright (C) 2019  Gully Chen
# MIT License
"""
setup file
"""

from pyzil import version
from setuptools import setup, find_packages

packages = find_packages()
package_data = {"pyzil": ["tests/*", "tests/crypto/*"]}

tests_require = ["pytest"]
install_requires = [
    "requests", "jsonrpcclient", "jsonrpcclient[requests]",
    "protobuf", "fastecdsa", "pyethash",
    "pycryptodome", "eth-hash[pycryptodome]",
]

setup(
    name="pyzil",
    version=version,
    description="Zilliqa Python API",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
    ],
    keywords="zilliqa api",
    author="Gully Chen",
    author_email="deepgully@gmail.com",
    url="http://github.com/deepgully/pyzil",
    license="MIT License",
    packages=packages,
    include_package_data=True,
    package_data=package_data,
    install_requires=install_requires,
    tests_require=tests_require,
)
